package v1

import (
	"context"
	"errors"
	"fmt"
	"git.getensh.com/common/gopkgs/cache"
	"git.getensh.com/common/gopkgs/logger"
	"github.com/gin-gonic/gin"
	"github.com/go-pay/gopay"
	"github.com/go-pay/gopay/wechat"
	wechatv3 "github.com/go-pay/gopay/wechat/v3"
	"go.uber.org/zap"
	"google.golang.org/grpc/status"
	"net/http"
	"property-callback-gateway/parser"
	"property-callback-gateway/pb"
	pb_v1 "property-callback-gateway/pb/v1"
	"strconv"
	"strings"
	"time"
)

type OrderPayInfo struct {
	PayTime int64
	OrderNo string
	PayFee  uint32
	Transid string
	// 1 小程序支付
	Type int
}

func GetOrderPayInfo(bodyMap gopay.BodyMap) (*OrderPayInfo, error) {

	info := &OrderPayInfo{Type: 1}
	info.Transid = bodyMap.Get("transaction_id")
	if info.Transid == "" {
		return nil, status.Error(10006, "缺少交易号")
	}
	info.OrderNo = bodyMap.Get("out_trade_no")
	if info.OrderNo == "" {
		return nil, status.Error(10006, "缺少订单号")
	}
	//todo 测试
	str := bodyMap.Get("cash_fee")
	if str == "" {
		return nil, status.Error(10006, "缺少金额")
	}
	integer, _ := strconv.Atoi(str)
	info.PayFee = uint32(integer)

	timeEnd := bodyMap.Get("time_end")
	if timeEnd == "" {
		return nil, status.Error(10006, "缺少缴费时间")
	}

	if ptime, err := time.ParseInLocation("20060102150405", timeEnd, time.Local); err != nil {
		return nil, err
	} else {
		info.PayTime = ptime.Unix()
	}

	return info, nil
}

//
// @Summary 微信小程序支付回调
// @Description 微信小程序支付回调
// @Tags callback
// @Accept  json
// @Produce  json
// @Router /api/v1/wx/applet/pay_callback [post]
func (c *Controller) WxMiniPayCallback(ctx *gin.Context) {
	ok := false
	var err error
	var payInfo *OrderPayInfo
	var bodyMap gopay.BodyMap

	defer func() {
		if r := recover(); r != nil {
			fmt.Printf("WxOrderPayCallback failed:%v\n", r)
			logger.Error("func",
				zap.String("call", "WxOrderPayCallback panic"),
				zap.String("error", fmt.Sprintf("%v", r)))
			ctx.String(http.StatusBadGateway, string("<xml><return_code><![CDATA[FAIL]]></return_code><return_msg><![CDATA[FAIL]]></return_msg></xml>"))
			return
		}
		if err != nil {

			logger.Error("func",
				zap.String("call", "WxOrderPayCallback panic"),
				zap.String("error", err.Error()))
			fmt.Printf("WxOrderPayCallback failed:%v\n", err)
			ctx.String(http.StatusBadGateway, string("<xml><return_code><![CDATA[FAIL]]></return_code><return_msg><![CDATA[FAIL]]></return_msg></xml>"))
		} else {
			fmt.Printf("WxOrderPayCallback success\n")
			ctx.String(http.StatusOK, string("<xml><return_code><![CDATA[SUCCESS]]></return_code><return_msg><![CDATA[OK]]></return_msg></xml>"))
		}
	}()

	// 解析参数
	bodyMap, err = wechat.ParseNotifyToBodyMap(ctx.Request)
	if err != nil {
		return
	}
	fmt.Printf("body map:%v\n", bodyMap)

	ok, err = wechat.VerifySign(parser.Conf.ThirdParty.Wx.AppletAppKey, wechat.SignType_MD5, bodyMap)
	if err != nil {
		return
	}
	if !ok {
		err = status.Error(10006, "小程序验签未通过")
		return
	}
	if parser.Conf.PayTest {
		return
	}

	payInfo, err = GetOrderPayInfo(bodyMap)
	if err != nil {
		return
	}

	resultKey := fmt.Sprintf("wxresult_%s_%s", payInfo.Transid, payInfo.OrderNo)
	result, _ := cache.Redis().Get(resultKey)
	if result == "success" {
		return
	}

	array := strings.Split(payInfo.OrderNo, "-")
	if len(array) < 2 {
		return
	}
	if len(array[0]) < 2 {
		return
	}
	gardenId, _ := strconv.ParseInt(array[0][1:], 10, 64)
	mreq := pb_v1.ChargeOrderPayRequest{
		OrderId:       payInfo.OrderNo,
		GardenId:      gardenId,
		TransactionId: payInfo.Transid,
		PayTime:       payInfo.PayTime,
		Amount:        int64(payInfo.PayFee),
	}
	fmt.Printf("xxx:%v\n", mreq)
	_, err = pb.Garden.ChargeOrderPay(context.Background(), &mreq)
	if err != nil {
		return
	}
	cache.Redis().SetEx(resultKey, 5, "success")
}

type V3NotifyResponse struct {
	Code    string `json:"code"`
	Message string `json:"message"`
}

//
// @Summary 微信小程序支付回调
// @Description 微信小程序支付回调
// @Tags callback
// @Accept  json
// @Produce  json
// @Router /apis/v1/wx/applet/pay_callbackv3 [post]
func (c *Controller) WxMiniPayCallbackv3(ctx *gin.Context) {
	var err error
	var v3req *wechatv3.V3NotifyReq

	defer func() {
		if r := recover(); r != nil {
			fmt.Printf("WxOrderPayCallback failed:%v\n", r)
			logger.Error("func",
				zap.String("call", "WxOrderPayCallback panic"),
				zap.String("error", fmt.Sprintf("%v", r)))
			ctx.JSON(http.StatusBadGateway, &V3NotifyResponse{Code: "FAIL", Message: "panic"})
			return
		}
		if err != nil {

			logger.Error("func",
				zap.String("call", "WxOrderPayCallback panic"),
				zap.String("error", err.Error()))
			fmt.Printf("WxOrderPayCallback failed:%v\n", err)
			ctx.JSON(http.StatusBadGateway, &V3NotifyResponse{Code: "FAIL", Message: "panic:" + err.Error()})
		} else {
			fmt.Printf("WxOrderPayCallback success\n")
			ctx.String(http.StatusOK, "")
		}
	}()

	v3req, err = wechatv3.V3ParseNotify(ctx.Request)
	if err != nil {
		fmt.Printf("parse v3 notify error:%v\n", err)
		return
	}

	mchID := parser.Conf.ThirdParty.Wx.MerchantMchID
	mchCertificateSerialNumber := parser.Conf.ThirdParty.Wx.MerchantMchCertificateSerialNumber
	mchAPIv3Key := parser.Conf.ThirdParty.Wx.MerchantMchAPIv3Key

	clientv3, er := wechatv3.NewClientV3(mchID, mchCertificateSerialNumber, mchAPIv3Key, parser.Conf.ThirdParty.Wx.MerchantMchPrivateKey)
	if er != nil {
		err = er
		fmt.Printf("new client v3 err:%v\n", err)
		return
	}
	certs, er := clientv3.GetPlatformCerts()
	if er != nil {
		err = er
		fmt.Printf("GetPlatformCerts err:%v\n", err)
		return
	}
	if len(certs.Certs) == 0 {
		fmt.Printf("GetPlatformCerts err:证书为空\n")
		err = errors.New("GetPlatformCerts err:证书为空")
		return
	}

	err = v3req.VerifySign(certs.Certs[0].PublicKey)
	if err != nil {
		fmt.Printf("验证签名失败\n")
		return
	}

	dresult, er := v3req.DecryptCipherText(mchAPIv3Key)
	if er != nil {
		err = er
		fmt.Printf("解密回调失败：%v\n", er)
		return
	}
	fmt.Printf("xxxxxxxxxxxxxx sucesss\n")

	// 解析参数
	if strings.Contains(dresult.OutTradeNo, "pay tests") {
		return
	}

	resultKey := fmt.Sprintf("wxresultv3_%s_%s", dresult.TransactionId, dresult.OutTradeNo)
	result, _ := cache.Redis().Get(resultKey)
	if result == "success" {
		return
	}

	array := strings.Split(dresult.OutTradeNo, "-")
	if len(array) < 2 {
		return
	}
	if len(array[0]) < 2 {
		return
	}

	tarray := strings.Split(dresult.SuccessTime, "+")
	payTime, er := time.ParseInLocation("2006-01-02 15:04:05", strings.ReplaceAll(tarray[0], "T", " "), time.Local)
	if er != nil {
		fmt.Printf("时间错误：%v，%s\n", er, dresult.SuccessTime)
		payTime = time.Now()
	}
	gardenId, _ := strconv.ParseInt(array[0][1:], 10, 64)
	mreq := pb_v1.ChargeOrderPayRequest{
		OrderId:       dresult.OutTradeNo,
		GardenId:      gardenId,
		TransactionId: dresult.TransactionId,
		PayTime:       payTime.Unix(),
		Amount:        int64(dresult.Amount.Total),
	}
	_, err = pb.Garden.ChargeOrderPay(context.Background(), &mreq)
	if err != nil {
		return
	}
	fmt.Printf("all success\n")
	cache.Redis().SetEx(resultKey, 5, "success")
}
